import typer
import clipboard
from typing import Annotated
from pathlib import Path
from rich.console import Console
from rich.syntax import Syntax
from rich.text import Text
from pygments.styles import get_all_styles

from .custom_typer import Typer
from .chains import create_chain, CodeWhisper
from .config import app as config_app
from .settings import config, UserConfig


app = Typer()


@app.command(default=True)
def ask(
    prompt: Annotated[str, typer.Argument(help="The question you want to ask Whisper")],
    conf: Annotated[
        str,
        typer.Option(
            "-c", "--config", help="The config to be used (openai, anthropic, etc)"
        ),
    ] = None,
    model: Annotated[
        str, typer.Option("-m", "--model", help="The model that Whisper should use")
    ] = None,
    theme: Annotated[
        str,
        typer.Option(
            "-t", "--theme", help="The theme that should be used for the output"
        ),
    ] = "solarized-dark",
    copy: Annotated[
        bool, typer.Option("--copy/--no-copy", help="Copy the output to the clipboard")
    ] = None,
):
    """
    The default command that takes a prompt and optional parameters to ask a question using Whisper.
    """
    console = Console(color_system="auto")

    with console.status("Thinking...", spinner="dots"):
        chain = create_chain(conf, model)
        result = chain.invoke({"context": prompt})

    if copy is not None:
        if copy and result.snippet:
            clipboard.copy(result.snippet)
    elif config.default.copy and result.content:
        clipboard.copy(result.content)

    panel = result.render(theme)
    console.print(panel)


@app.command()
def themes():
    """Prints a list of available syntax highlighting themes from Pygments."""
    console = Console()
    console.print(Text(", ".join(list(get_all_styles())), style="bold green"))


@app.command()
def init(
    openai_key: str = typer.Option(
        None, "--openai-key", help="The OpenAI API key to use"
    ),
    anthropic_key: str = typer.Option(
        None, "--anthropic-key", help="The Anthropic API key to use"
    ),
    azureopenai_key: str = typer.Option(
        None, "--azureopenai-key", help="The Azure OpenAI API key to use"
    ),
    mistral_key: str = typer.Option(
        None, "--mistral-key", help="The Mistral API key to use"
    ),
    fireworks_key: str = typer.Option(
        None, "--fireworks-key", help="The Fireworks API key to use"
    ),
    show: Annotated[
        bool, typer.Option("--show/--no-show", help="Show the configuration")
    ] = True,
):
    """
    Initializes the configuration for Whisper, saving API keys to a local file.
    """
    console = Console()
    config_path = Path.home() / ".whisper"
    config_path.mkdir(parents=True, exist_ok=True)
    config_file = config_path / "whisper.toml"

    with config_file.open("w") as f:
        toml_str = UserConfig.initialize(
            openai_key, anthropic_key, mistral_key, fireworks_key, azureopenai_key
        )
        f.write(toml_str)

    if show:
        syntax = Syntax(toml_str, "toml", theme=config.default.theme, padding=1)
        panel = CodeWhisper.render_syntax(syntax, "Whisper Configuration")
        console.print(panel)

    console.print(f"[green]Initialized whisper configuration in {config_file}[/green]")


app.add_typer(config_app, name="config")

if __name__ == "__main__":
    app()



import subprocess
import markdown
import xml.etree.ElementTree as etree

from mkdocs.plugins import BasePlugin
from mkdocs.config import config_options
from markdown.inlinepatterns import InlineProcessor


class TyperExtension(markdown.Extension):
    def __init__(self, *args, cmd: str = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.cmd = cmd

    def extendMarkdown(self, md):
        # Register the new pattern to handle ::typer
        md.inlinePatterns.register(
            TyperPattern(r"::typer", md, cmd=self.cmd), "typer", 175
        )


class TyperPattern(InlineProcessor):
    def __init__(self, *args, cmd: str = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.cmd = cmd

    def handleMatch(self, m, data):
        result = subprocess.run(self.cmd.split(" "), capture_output=True, text=True)

        if result.returncode == 0:
            html_output = markdown.markdown(result.stdout)
            return etree.fromstring(f"<div>{html_output}</div>"), m.start(0), m.end(0)
        else:
            return None, m.start(0), m.end(0)


def makeExtension(**kwargs):
    return TyperExtension(**kwargs)


class MkdocsTyper(BasePlugin):
    config_scheme = (
        (
            "cmd",
            config_options.Type(str, default="typer whisper utils docs --name Whisper"),
        ),
    )

    def on_config(self, config):
        """Called after the configuration is loaded."""
        # Add the updated Markdown extension to the list
        config["markdown_extensions"].append(makeExtension(cmd=self.config["cmd"]))
        return config

    def on_pre_build(self, config):
        """Called before the MkDocs build process starts."""
        pass



from inspect import isclass
import toml

from pathlib import Path
from typing import Type, Tuple

from pydantic_core import PydanticUndefinedType, ValidationError
from pydantic import BaseModel, Field
from rich.console import Console

from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)


class BaseModelConfig(BaseModel):
    api_key: str = "replace-me"
    model: str


class OpenAIConfig(BaseModelConfig):
    model: str = "gpt-3.5-turbo"


class AnthropicConfig(BaseModelConfig):
    model: str = "claude-3-5-sonnet-20240620"


class MistralAIConfig(BaseModelConfig):
    model: str = "pixtral-12b-2409"


class FireworksConfig(BaseModelConfig):
    model: str = "accounts/fireworks/models/llama-v3p1-405b-instruct"


class AzureOpenAIConfig(BaseModelConfig):
    model: str = "gpt-4o"


class DefaultConfig(BaseModel):
    copy_snippet: bool | None = None
    config: str = "openai"
    theme: str | None = "solarized-dark"


class UserConfig(BaseSettings):
    model_config = SettingsConfigDict(toml_file=["~/.whisper/whisper.toml"])

    default: DefaultConfig = Field(default_factory=DefaultConfig)

    openai: OpenAIConfig = Field(default_factory=OpenAIConfig)
    anthropic: AnthropicConfig = Field(default_factory=AnthropicConfig)
    mistral: MistralAIConfig = Field(default_factory=MistralAIConfig)
    fireworks: FireworksConfig = Field(default_factory=FireworksConfig)
    azureopenai: AzureOpenAIConfig = Field(default_factory=AzureOpenAIConfig)

    def get_by_dot_notation(self, key: str):
        keys = key.split(".")
        value = self
        for k in keys:
            value = getattr(value, k)
        return value

    def set_by_dot_notation(self, key: str, value: str):
        keys = key.split(".")
        v = self
        for k in keys[:-1]:
            v = getattr(v, k)
        setattr(v, keys[-1], value)

    def save(self):
        path = Path.home() / ".whisper" / "whisper.toml"
        with path.open("w") as f:
            toml.dump(self.model_dump(), f)

    @classmethod
    def config_exists(cls):
        for path in cls.model_config["toml_file"]:
            if Path(path).exists():
                return True
        return False

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (TomlConfigSettingsSource(settings_cls),)

    @classmethod
    def initialize(
        cls,
        openai_key: str = None,
        anthropic_key: str = None,
        mistral_key: str = None,
        fireworks_key: str = None,
        azureopenai_key: str = None,
    ) -> str:
        def recurse_fields(cls_fields):
            data = {}
            for key, field in cls_fields.items():
                if isclass(field.annotation) and issubclass(
                    field.annotation, BaseModel
                ):
                    data[key] = recurse_fields(field.annotation.model_fields)
                elif not isinstance(field.default, PydanticUndefinedType):
                    data[key] = field.default
            return data

        config_data = recurse_fields(cls.model_fields)

        if openai_key is not None:
            config_data["openai"]["api_key"] = openai_key
        if anthropic_key is not None:
            config_data["anthropic"]["api_key"] = anthropic_key
        if mistral_key is not None:
            config_data["mistral"]["api_key"] = mistral_key
        if fireworks_key is not None:
            config_data["fireworks"]["api_key"] = fireworks_key
        if azureopenai_key is not None:
            config_data["azureopenai"]["api_key"] = azureopenai_key

        return toml.dumps(config_data)


try:
    config = UserConfig()
except ValidationError as e:
    console = Console()
    console.print(f"[red]Error: {e}[/red]")

    if not UserConfig.config_exists():
        console.print(
            "[yellow]No configuration file found.  Please run `whisper init` to initialize your configuration.[/yellow]"
        )
    else:
        console.print(
            "[yellow]It looks like your configuration file is malformed.  Please fix the errors and try again.[/yellow]"
        )
except Exception as e:
    console = Console()
    console.print(f"[red]Error: {e}[/red]")



from pydantic import BaseModel, Field, field_validator, ValidationInfo
from langchain_openai import ChatOpenAI, AzureChatOpenAI
from langchain_anthropic import ChatAnthropic
from langchain_fireworks import ChatFireworks
from langchain_mistralai import ChatMistralAI
from langchain_core.prompts import ChatPromptTemplate
from rich.panel import Panel
from rich.markdown import Markdown
from rich.syntax import Syntax


from .settings import UserConfig

config = UserConfig()


TEMPLATE = ChatPromptTemplate.from_template(
    """
You are an AI assistant that can answer questions and help with tasks.

{context}
"""
)


class CodeWhisper(BaseModel):
    is_prompt_code_related: bool = Field(
        description="Whether or not the prompt / question was code or command related"
    )
    title: str = Field(description="A short title for the AI's response")
    content: str = Field(
        description="The AI's response to the prompt in Markdown format."
    )
    language: str | None = Field(
        description="The programming language of the code snippet or terminal command"
    )
    snippet: str | None = Field(
        description="The code snippet or terminal command if the prompt was code related"
    )

    @field_validator("language", "snippet", mode="before")
    @classmethod
    def check_code_related_fields(cls, v: str, info: ValidationInfo):
        if info.data.get("is_prompt_code_related") and not v:
            raise ValueError(
                f"The [{info.field_name}] must be provided if the prompt was code related"
            )
        return v

    def render(self: "CodeWhisper", theme: str = "solarized-dark") -> Panel:
        return Panel(
            Markdown(self.content, code_theme=theme),
            title=self.title or "Result",
            border_style="green",
        )

    @staticmethod
    def render_syntax(
        syntax: Syntax,
        title: str,
        border_style: str = "green",
    ) -> Panel:
        return Panel(
            syntax,
            title=title,
            border_style=border_style,
        )


def create_chain(key: str = None, model: str = None):
    key = key or config.default.config
    model = model or getattr(config, key).model

    if key == "openai":
        return TEMPLATE | ChatOpenAI(
            model=model, temperature=0, openai_api_key=config.openai.api_key
        ).with_structured_output(CodeWhisper)
    elif key == "anthropic":
        return TEMPLATE | ChatAnthropic(
            model=model, temperature=0, anthropic_api_key=config.anthropic.api_key
        ).with_structured_output(CodeWhisper)
    elif key == "mistral":
        return TEMPLATE | ChatMistralAI(
            model=model, temperature=0, mistral_api_key=config.mistral.api_key
        ).with_structured_output(CodeWhisper)
    elif key == "fireworks":
        return TEMPLATE | ChatFireworks(
            model=model, temperature=0, fireworks_api_key=config.fireworks.api_key
        ).with_structured_output(CodeWhisper)
    elif key == "azureopenai":
        return TEMPLATE | AzureChatOpenAI(
            model=model, temperature=0, azure_api_key=config.azureopenai.api_key
        ).with_structured_output(CodeWhisper)
    else:
        raise ValueError(f"Key {key} not found")
